/* conffile.c
 *
 * Generic Config Tools
 *
 * Copyright (C) 2003 Christopher Collins <ccollins@ug.cs.usyd.edu.au>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *                                                                             
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *                                                                             
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/*
 * basic config file format is:
 *
 * ---BEGIN---
 * # comment
 * ; comment
 * key = value
 * ---END---
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <assert.h>
#include <errno.h>

#include "strutil.h"
#include "conffile.h"

extern int
config_parsevalue(char* strin, config_type_t type, config_value_t* cfgval)
{
  long int	lval;
  double	fval;
  char*		eptr;  

  switch (type) {
  case ct_boolean:
    switch (strin[0]) {
      case 'T': case 't':
      case 'Y': case 'y':
        cfgval->bool = 1;
        break;
      case 'F': case 'f':
      case 'N': case 'n':
        cfgval->bool = 0;
        break;
      default:
        return (-1);
      }
    return (0);
  case ct_string:
    if (NULL != cfgval->string) {
      free (cfgval->string);
    }
    cfgval->string = strdup(strin);
    return (0);
  case ct_integer:
    lval = strtol(strin, &eptr, 10);
    if (eptr != strin+strlen(strin)) {
      return (-1);
    }
    cfgval->integer = lval;
    return (0);
  case ct_float:
    fval = strtod(strin, &eptr);
    if (eptr != strin+strlen(strin)) {
      return (-1);
    }
    cfgval->fp = (float)fval;
    return (0);
  default:
    return (-1);
  }
}


extern int
config_set(config_t* root, char* key, char* value)
{
	config_t*	tptr;
	
	/* if these fail, then something is broken */
	assert(key != NULL);
	assert(value != NULL);
	/* assert lowercase on the key */
	strlcase(key);
	/* search to see if the key already exists */
	tptr = root;
	while (tptr != NULL) {
		int insloc = strcmp(key, tptr->key);
		if (!insloc) {
			break;
		} else if (insloc > 0) {
			/* right branch */
			tptr = tptr->right;
		} else {
			/* left branch */
			tptr = tptr->left;
		}
	}
  if (tptr != NULL) {
    if (config_parsevalue(value, tptr->type, &tptr->value)) {
      fprintf (stderr, "Tried to set key \"%s\" to invalid value \"%s\"\n", key, value);
      return (-1);
    }
	} else {
    /* this is wrong! */
    fprintf (stderr, "Tried to set unknown key \"%s\" to \"%s\"\n", key, value);
    return (-1);
  }
	return (0);
}

extern config_t*
config_bind(config_t* root, char* key, char* value, config_type_t type)
{
	config_t*	tptr;
	config_t*	newnode;
	int 		insloc = 0;
	
	/* if these fail, then something is broken */
	assert(key != NULL);
	assert(value != NULL);
	/* assert lowercase on the key */
	strlcase(key);
	/* search to see if the key already exists */
	tptr = root;
	while (tptr != NULL) {
		insloc = strcmp(key, tptr->key);
		if (!insloc) {
			break;
		} else if (insloc > 0) {
			/* right branch */
			if (NULL == tptr->right) {
				break;
			}
			tptr = tptr->right;
		} else {
			/* left branch */
			if (NULL == tptr->left) {
				break;
			}
			tptr = tptr->left;
		}
	}
	if (insloc == 0 && root != NULL) {
		/* replace node */
    /* node occupied! */
    fprintf (stderr, "Tried to rebind node \"%s\" with value \"%s\"\n", key, value);
    return (root);
	} else {
		/* build a key, and insert */
		newnode = malloc(sizeof(*newnode));
		if (NULL == newnode) {
			/* b0rkage! */
			return (NULL);
		}
		memset(newnode, 0, sizeof (*newnode));
		newnode->key = strdup(key);
    newnode->type = type;
    config_parsevalue(value, type, &newnode->value);
    /*
		newnode->value = strdup(value);
    */
		if (0 == insloc) {
			/* rootnode */
			root = newnode;
		} else if (insloc > 0) {
			tptr->right = newnode;
		} else {
			tptr->left = newnode;
		}
	}
	return (root);
}


extern int
config_read (config_t* root, char* filename)
{
	FILE*		fin = NULL;
	char		buffer[2048];
	char*		rptr;
	
	fin = fopen(filename, "rt");
	if (NULL == fin) {
		return (-1);
	}
	fgets(buffer, 2048, fin);
	while (!feof(fin) && !ferror(fin)) {
		/* check for blank lines */
		strtrim(buffer);
		if ('\0' == buffer[0]) {
			goto nextloop;
		}
		/* check for comments, etc */
		if ('#' == buffer[0] || ';' == buffer[0]) {
			goto nextloop;
		}
		/* split the string */
		rptr = strchr (buffer, '=');
		*(rptr++) = '\0';
		/* clean up halves */
		strtrim(rptr);
		strtrim(buffer);

		config_set(root, buffer, rptr);	       
	  nextloop:
		fgets(buffer, 2048, fin);
	}
	fclose (fin);

	return (0);
}

extern void
config_free(config_t* root)
{
	if (NULL == root)
		return;
	if (root->left != NULL) {
		config_free(root->left);
		root->left = NULL;
	}
	if (root->right != NULL) {
		config_free(root->right);
		root->right = NULL;
	}
  if (root->type == ct_string && NULL != root->value.string) {
    free (root->value.string);
  }
	free (root);
}

extern config_t*
config_find(config_t* root, char* key)
{
	config_t*	tptr;
	int		insloc = 0;

	tptr = root;
	while (tptr != NULL) {
		insloc = strcmp(key, tptr->key);
		if (!insloc) {
			break;
		} else if (insloc > 0) {
			/* right branch */
			tptr = tptr->right;
		} else {
			/* left branch */
			tptr = tptr->left;
		}
	}
	return (tptr);
}

extern void
config_print(config_t *root, FILE *fout)
{
	if (NULL == root)	
		return;
	if (NULL != root->left)	
		config_print(root->left, fout);
  switch (root->type) {
  case ct_boolean:
    if (root->value.bool) {
      fprintf(fout, "%s = true\n", root->key);
    } else {
      fprintf(fout, "%s = false\n", root->key);
    }
    break;
  case ct_integer:
    fprintf(fout, "%s = %ld\n", root->key, root->value.integer);
    break;
  case ct_float:
    fprintf(fout, "%s = %0.12f\n", root->key, root->value.fp);
    break;
  case ct_string:
    if (root->value.string != NULL) {
      fprintf(fout, "%s = %s\n", root->key, root->value.string);
    }
    break;
  default:
    break;
  }
	if (NULL != root->right)
		config_print(root->right, fout);
}
