/* fah.c
 *
 * FAH related routines.
 *
 * Copyright (C) 2003 Christopher Collins <ccollins@ug.cs.usyd.edu.au>
 * Copyright (C) 2002 Richard P. Howell IV.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *                                                                             
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *                                                                             
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
*/

#include "xfpd.h"

#define MAXLEN_PNAME	32

char *		xyzname = NULL;
char *		defaultdatafile = NULL;
char		pname[MAXLEN_PNAME];	/* Name of protein */

#define BUFSIZE		200

/* deletes the bond specified */
static void
trim_bond(struct atm *srcatm, int bondnum)
{
	/* bonus bond */
	if (srcatm->bondcount <= bondnum) {
		return;
	}
	/* check if a shuffle is necessary */
	if ((srcatm->bondcount - bondnum) > 1) {
		/* shuffle bonds back one. */
		memmove(&srcatm->bond[bondnum], 
			&srcatm->bond[bondnum+1], 
			sizeof(*(srcatm->bond)) * (srcatm->bondcount - bondnum - 1));
	}
	/* reduce the boundcount by one. */
	srcatm->bondcount--;

	/* and we're done! */
	return;
}

/* Calculate bond length, theta, and phi */
void bondparm(struct atm *p1, struct atm *p2, struct bnd *pb)
{
	double x, y, z, r, t;

	/* calculate r, the distance between the two atoms */
	x = p2->pos[V_X] - p1->pos[V_X];
	y = p2->pos[V_Y] - p1->pos[V_Y];
	z = p2->pos[V_Z] - p1->pos[V_Z];
	r = sqrt(x * x + y * y + z * z);

	t = (x == 0.0) ? 90.0 : acos(x / sqrt(x * x + y * y)) * 180.0/M_PI;
	pb->len = r / 2.0;
	pb->theta = (y < 0.0) ? -t: t;
	pb->phi = (r == 0.0) ? 0.0 : acos(z / r) * 180.0/M_PI;
}

/* quickly allocate an ach pair, and initialise them */
int
alloc_ach(struct atm *p)
{
	p->achc = malloc(sizeof(*(p->achc)));
	p->acha = malloc(sizeof(*(p->acha)));
	if (NULL == p->acha || NULL == p->acha) {
		return -1;
	}
	memset(p->acha, 0, sizeof(*(p->acha)));
	memset(p->achc, 0, sizeof(*(p->achc)));

	p->acha->toward = -1;
	p->achc->toward = -1;

	return 0;
}

/* Find peptide bonds and set up the alpha carbon chain */
int 
achain(void)
{
	int i, j;
	int n0, n1, n2;
	int b0, b1;
//	struct atm *p0, *p1, *p2;
	char *p;

	p = ETYPES;

	/* Step 1: find all carbon - carbonyl - imino sequences */
	for (n0 = 0; n0 < atoms; n0++) {
		i = p[atom[n0].element];

		if (i == 'H')
			atom[n0].flags |= AF_H;
		if (i != 'C')
			continue;

		for (b0 = 0; b0 < atom[n0].bondcount; b0++) {
		//for (b0 = MAXBONDS; --b0 >= 0; )
			n1 = atom[n0].bond[b0].toward;
			if (p[atom[n1].element] != 'C') {
				continue;
			}

			i = 0; 
			j = 0;

			for (b1 = 0; b1 < atom[n1].bondcount; b1++) {
				n2 = atom[n1].bond[b1].toward;
				/* can't loop backwards */
				if (n2 == n0) {
					continue;
				}

				if (p[atom[n2].element] == 'O') {
					++i;
				} else if (p[atom[n2].element] == 'N') {
					j = n2;
					i += 10;
				} else i += 1000;
			}
			if (i == 11) {
				if (alloc_ach(&atom[n0]))
					return -1;
				atom[n0].achc->toward = j;		/* Temporary pointer from C to N */
			} else if ((i == 1) || (i == 2)) {
				if (alloc_ach(&atom[n0]))
					return -1;

				atom[n0].achc->toward = -1;		/* At least eligible to be alpha C */
			}
		}
	}

	/* Step 2: find imino - alpha-carbon sequences and peptide bonds */

	for (n0 = 0; n0 < atoms; n0++) {
		if (NULL == atom[n0].achc)
			continue;		
		n1 = atom[n0].achc->toward;
		if (n1 < 0)
			continue;
		atom[n0].achc->toward = -1;
		for (b1 = 0; b1 < atom[n1].bondcount; b1++) {
			n2 = atom[n1].bond[b1].toward;
			/* if the bond leads to a elegable atom, then link it and continue */
			if (NULL != atom[n2].achc) {
				atom[n0].achc->toward = n2;			/* Peptide bond, carbonyl side */
				atom[n2].acha->toward = n0;			/* Peptide bond, imino side */
				break;
			}
		}
	}

	/* Step 3: calculate angles and flush isolated peptide bonds */
	for (n0 = 0; n0 < atoms; n0++) {
		if (atom[n0].achc == NULL || atom[n0].achc->toward < 0)
			continue;
		n2 = atom[n0].achc->toward;
		if (atom[n0].acha != NULL 
			&& atom[n2].achc != NULL) {
			bondparm(&atom[n0], &atom[n2], atom[n0].achc);
			bondparm(&atom[n2], &atom[n0], atom[n2].acha);
		} else {
			atom[n2].acha->toward = -1;
			atom[n0].achc->toward = -1;
		}
	}

	/* Extra step: identify water molecules */
	/* XF:  Because we now know how many live bonds there are,
	 *      this classification test is SO MUCH SIMPLER! */
	for (n0 = 1; n0 <= atoms; ++n0) {
		if (p[atom[n0].element] != 'O' 
			|| atom[n0].bondcount != 2) {
			continue;
		}
		for (b0 = 0; b0 < atom[n0].bondcount; b0++) {
			n1 = atom[n0].bond[b0].toward;
			if (!(atom[n1].flags & AF_H) 
				||  atom[n1].bondcount != 1 
				|| atom[n1].bond[0].toward != n0) {
				break;
			}
		}
		if (b0 != atom[n0].bondcount) {
			continue;
		}
		atom[n0].flags |= AF_H2O;  /* Mark the Oxygen */
		for (b0 = 0; b0 < atom[n0].bondcount; b0++) {
			atom[atom[n0].bond[b0].toward].flags |= AF_H2O;
		}
	}

	/* ALL OK */
	return 0;
}

/* Clean up bond data */
void xyzclean()
{
	int		i;
	int		n0, n1;
	int		b0, b1;
//	struct atm *	p0;
//	struct atm *	p1;

	/* Step 1: identify XYZ file type (first "bond" pointer is
	 * atom ID with Genome and Tinker) */
	i = atoms;
	for (n0 = 0; n0 < atoms; n0++) {
		if (atom[n0].bondcount <= 0 ||
			atom[n0].bond[0].toward > atoms) {
			/* no bonds, or first bond invalid. */
			continue;
		}
		n1 = atom[n0].bond[0].toward;
		for (b1 = 0; b1 < atom[n1].bondcount; b1++) {
			if (atom[n1].bond[b1].toward == n0) {
				--i;
				break;
			}
		}
	}
	/* logic:  if over 1/5th of the first bonds are non-reciprocal,
	 * then we must be using an odd format */
	if (atoms <= 5*i) {
		for (n0 = 0; n0 < atoms; n0++) {
			trim_bond(&atom[n0], 0);
		}
	}

	/* Step 2: clean up bad bond pointers */
	for (n0 = 0; n0 < atoms; n0++) {
		for (b0 = 0; b0 < atom[n0].bondcount; b0++) {
			n1 = atom[n0].bond[b0].toward;
			if (n1 == n0) {
				/* bond is self referential */
				trim_bond(&atom[n0], b0);
				b0--;
				continue;
			}
			if (n1 > atoms)	{
				/* bond out of range */
				trim_bond(&atom[n0], b0);
				b0--;
				continue;
			}
			for (i = b1 = 0; b1 < atom[n1].bondcount; b1++) {
				if (atom[n1].bond[b1].toward == n0) {
					i++;
					if (i > 1) {
						trim_bond(&atom[n1], b1);
						b1--;
						continue;
					}
				}
			}
			if (i == 0)
			{
				/* Not reciprocal */
				trim_bond(&atom[n0], b0);
				b0--;
				continue;
			}
		}
	}
}

#define CH_BOND_LEN		1.095
#define CH_BOND_ID		21

/* Read in the XYZ file data for the molecule to display */
extern int 
readxyz ()
{
	int		i, j, k, n;
	char *		p;
	char *		q;
	FILE *		fp;
//	struct atm *	pa;
	int		na;
	double		ascale[3], bscale[3];
//	double		xa, xb, ya, yb, za, zb;
	double		chdst;
	double		d, g;
	struct stat	st;
	int		retval = 0;
	char		wbuf[BUFSIZE];
	char		ebuf[8];

	static int	xyztime = 0;
	char *		fn = NULL;

	p = ETYPES;

	if (NULL == xyzname) {
		fn = defaultdatafile;
	} else {
		fn = xyzname;
	}

	if (stat(fn, &st) != 0)
		return (FAH_CANTOPEN);

	if (st.st_mtime == xyztime)
		return (0);	/* File hasn't changed since last time */

	if ((fp = fopen(fn, "r")) == NULL)
		return (FAH_CANTOPEN);

	na = 0;
	/* get the number of atoms and the name of the protein */
	if (fgets(wbuf, BUFSIZE, fp) != NULL) {		
		memset(pname, 0, MAXLEN_PNAME);
		sscanf(wbuf, "%d %" stringize(MAXLEN_PNAME) "c", &na, pname);
		for (i = 0; i < MAXLEN_PNAME; ++i)
			if (((j = pname[i]) == '\012') || (j == '\015')) {
				pname[i] = '\0';
				break;				
			}       
	}

	/* if the protein has no name... */
	if (pname[0] == '\0') {
		strcpy(pname, "Genome");
	}

	atoms = 0;
	++na;	/* Tinker reports one too few atoms */
	if (na >= MAXATOMS) {
		retval = FAH_TOOMANYATOMS;
		goto abort;
	}

	/* if there is still an old atom chain present, dispose of it */
	if (atom != NULL) {
		free(atom);
	}

	atom = calloc(na, sizeof(struct atm));
	if (NULL == atom) {
		retval = FAH_OUTOFMEMORY;
		goto abort;
	}

	while (fgets(wbuf, BUFSIZE, fp) != NULL && atoms < na) {
		int		zi, zn, zz;

		if (sscanf(wbuf, "%d%7s%f%f%f%n", &zi, ebuf, 
			&atom[atoms].pos[V_X], &atom[atoms].pos[V_Y], &atom[atoms].pos[V_Z], 
			&zn) < 5) {
			continue;
		}
		
		if (atoms+1 != zi) {
			retval = FAH_BADSEQUENCE;
			goto abort;
		}

		q = strchr(p, ebuf[0]);
		if (NULL == q) {
                        atom[atoms].element = 0;
		} else {
			atom[atoms].element = (q - p);
		}

		/* this uses zn to reference the start of the next arg in the string */
		for (zi = i = 0; sscanf(&wbuf[zn], "%d%n", &zz, &zi) == 1; zn += zi) {
			if (i >= MAXBONDS) {
				/* Too much garbage */
				retval = FAH_TOOMANYBONDS;
				goto abort;
			}
			/* bonds are now zero referenced */
			atom[atoms].bond[i++].toward = zz - 1; 
		}
		atom[atoms].bondcount = i;
		atoms++;
	}
abort:
	fclose(fp);
	if (retval)
		return (retval);

	if (atoms == 0)	{
		return (FAH_NOATOMS);
	}

	xyzclean();			/* Clean up bond data */

	g = 0.0;			/* Calculate scaling */

	for (i = 0; i < 3; i++) {
		ascale[i] = atom[0].pos[i];
		bscale[i] = atom[0].pos[i];
	}
	n = 0;
	chdst = 1e10;
	for (i = 1; i < atoms; i++) {
		for (j = 0; j < 3; j++) {
			if (ascale[j] > atom[i].pos[j])
				ascale[j] = atom[i].pos[j];
			if (bscale[j] < atom[i].pos[j])
				bscale[j] = atom[i].pos[j];
		}
		for (j = 0; j < atom[i].bondcount; j++) {
			k = atom[i].bond[j].toward;
			if (k <= 0)
				continue;
			if (atom[i].element * 10 + atom[k].element != CH_BOND_ID)
				continue;			/* Just look at lengths of C-H bonds */
			bondparm(&atom[i], &atom[k], &atom[i].bond[j]);	/* Temporarily calculate len */
			g += atom[i].bond[j].len;
			d = g * 1.2 / (double) ++n;
			if (d > 1.44 * chdst)
				g = chdst * (double) n;
			if ((chdst > d)
			    || ((atom[i].bond[j].len > chdst)
				&& (atom[i].bond[j].len < d))) {
				chdst = atom[i].bond[j].len;		/* Best guess for "typical" C-H bond */
			}
		}
	}
	if (chdst > 1e5) /* Not a reasonable value, set to default */
		chdst = CH_BOND_LEN / 2;

	for (i = 0; i < 3; i++) {
		ascale[i] = (ascale[i] + bscale[i]) / 2;
		bscale[i] -= ascale[i];
	}

	if (bscale[V_X] < bscale[V_Y]) {
		bscale[V_X] = bscale[V_Y];
	}
	if (bscale[V_X] < bscale[V_Z]) {
		bscale[V_X] = bscale[V_Z];
	}
	bscale[V_X] = 1 / bscale[V_X];
	scale = chdst * bscale[V_X] * 2 / CH_BOND_LEN;


	/* perform scaling to fit into the box */
	for (i = 0; i < atoms; i++) {
		for (j = 0; j < 3; j++) {
                        atom[i].pos[j] = (atom[i].pos[j] - ascale[j]) * bscale[V_X];
		}
	}

	/* do bond calculations */
	for (i = 0; i < atoms; i++) {
		for (j = 0; j < atom[i].bondcount; j++) {
			k = atom[i].bond[j].toward;
			bondparm(&atom[i], &atom[k], &atom[i].bond[j]);	/* Calculate len, theta, phi */
			d = atom[i].bond[j].len * 2.0 / scale;
			g = (aprops[atom[i].element].radius + aprops[atom[k].element].radius) * 0.63;
			if (d > g) {
				trim_bond(&atom[i], j); 	/* Something is wrong, bond is too long */
			}
		}
	}
	
	achain();		/* Find peptide bonds and set up the alpha carbon chain */

	xyztime = st.st_mtime;

	return (0);
}
