/* render.c
 *
 * Handles rendering of the FAH molecule 
 *
 * Copyright (C) 2003 Christopher Collins <ccollins@ug.cs.usyd.edu.au>
 * Copyright (C) 2002 Richard P. Howell IV.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *                                                                             
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *                                                                             
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "xfpd.h"

/** RENDERING PARAMETERS */

/* rendering system globals */
int		rendermode = DM_SF;

//XF: made obsolete by qrot.
GLfloat		qrot[4] = { 0.0, 0.0, 0.0, 0.0 };
double		dv[3] = { 0.0, 0.0, 1.0 };	/* huh? DV? */

/* zoom level */
double		zoom = IZOOM;

/* sphere rendering quality setting */
int		quality = 0;

/* aspect used for perspective.  external as it needs to be updated by glut */
double		aspect = 1.0;

/** MODEL PARAMETERS */
/* output flags */
int		oflags = 0;
/* size of balls for BS mode - range 0.0 - 1.0 */
double		bs_spheresize = 0.15;

/* Scale factor for normalizing molecule size - refernced in fah.c
*/
double		scale;		


int		atoms;	/* Total number of atoms in "atom" table */
struct atm *	atom;	/* Pointer to allocated storage for all XYZ file data */

//FIXME:  These are globals that should be localised.

static double	bond_radius;	/* Scaled bond thickness (displayed as cylinder) */


//FIXME: these don't belong in here, but I've nowhere else to put them for now.
#define FOV	45.0		/* Field of view */
#define DQUAL	5000.0		/* Medium display quality */

#define BOND_RADIUS		0.07
#define ALPHA_BOND_RADIUS	0.48
#define LARGEST_ATOM		2.10



/* Atomic radii, bond lengths, and similar things are in Angstroms.
** The display sizes of the atoms are, to some extent, a fudge.
*/

struct atmprop aprops[] = {
	{ 1.50, { 0.80, 0.10, 0.80, 1.00 } },	/* ? purple */
	{ 1.17, { 0.90, 0.90, 0.90, 1.00 } },	/* H white */
	{ 1.75, { 0.30, 0.30, 0.30, 1.00 } },	/* C dark grey */
	{ 1.55, { 0.20, 0.20, 0.80, 1.00 } },	/* N blue */
	{ 1.40, { 0.80, 0.10, 0.20, 1.00 } },	/* O red */
	{ 2.04, { 0.10, 0.60, 0.10, 1.00 } },	/* P green */
	{ 1.84, { 0.60, 0.60, 0.15, 1.00 } }	/* S yellow */
};

struct atmprop acprops[] = {		/* Pseudo-atom properties in alpha-carbon chain */
	{ 0.80, { 0.60, 0.65, 0.30, 1.00 } },	/* Normal gold */
	{ 0.80, { 0.15, 0.65, 0.60, 1.00 } },	/* Amino end turquoise */
	{ 0.80, { 0.70, 0.40, 0.40, 1.00 } },	/* Carboxyl end pink */
};

GLfloat bgcolor[3] = { 0.1, 0.1, 0.1 };


/* set up the GL view */
static void 
initView(GLUquadricObj **qobj)
{
	GLfloat front_specular[] = { 0.6, 0.6, 0.6, 1.0 };
	GLfloat ambient0[] = { 0.3, 0.3, 0.3, 1.0 };
	GLfloat diffuse0[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat position0[] = { 1.5, 1.5, 1.5, 0.0 };
	GLfloat ambient1[] = { 0.15, 0.15, 0.15, 1.0 };
	GLfloat diffuse1[] = { 0.5, 0.5, 0.5, 1.0 };
	GLfloat position1[] = { -1.5, -1.5, 1.5, 0.0 };
	GLfloat lmodel_ambient[] = { 0.5, 0.5, 0.5, 1.0 };
	GLfloat lmodel_twoside[] = { GL_TRUE };

	glColor4f(1.0, 1.0, 1.0, 1.0);

	glClearColor(bgcolor[0], bgcolor[1], bgcolor[2], 0.0);
	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);

	/* Lighting */
	if (rendermode != DM_FWF && rendermode != DM_FAT) {
		glLightfv(GL_LIGHT0, GL_AMBIENT, ambient0);
		glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse0);
		glLightfv(GL_LIGHT0, GL_POSITION, position0);
		glLightfv(GL_LIGHT1, GL_AMBIENT, ambient1);
		glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse1);
		glLightfv(GL_LIGHT1, GL_POSITION, position1);
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
		glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glEnable(GL_LIGHT1);

		glEnable(GL_NORMALIZE);
		glEnable(GL_CULL_FACE);
		glShadeModel(GL_SMOOTH);

		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_specular);
	} else {
		glDisable(GL_LIGHTING);
		glShadeModel(GL_FLAT);
	}


	if (NULL == *qobj) {
		*qobj = gluNewQuadric();
		gluQuadricDrawStyle(*qobj, GLU_FILL);
	}
}


/* render a bond */
static int 
dobond (GLUquadricObj *qobj, struct bnd *pb, int n)
{
	if ((oflags & OF_IH)	/* Don't display bonds to hydrogen */
		&& (atom[pb->toward].flags & AF_H))	/* and it's to hydrogen */
		return 0;

	glPushMatrix();
	glRotatef(pb->theta, 0, 0, 1);
	glRotatef(pb->phi, 0, 1, 0);
	gluCylinder(qobj, bond_radius, bond_radius, pb->len, n, 1);
	glPopMatrix();
	return (1);
}


/* Comparison function for sorting so we render nearest first */
//XF: Removed for now - display list sorting is rather unnecessary when you
// have depth checking.
//
//static int 
//cmpf(struct adist *pa, struct adist *pb)
//{
//	if (pa->r > pb->r) return (-1);
//	if (pa->r < pb->r) return (1);
//	return (0);
//}

/* Render the molecule */
static void 
drawMolecule(GLUquadricObj *qobj)
{
	int			i, j, n;
	struct atmprop *	pp;
	double			sscale, d, g, h;
	double			dqual;
	GLfloat			shiny;
	GLfloat			rotmatrix[4][4];
	struct adist *		patom;	/* Pointer to auxiliary pointer table for sorting */

	int			pectr;

	patom = calloc(atoms, sizeof(*patom));

	glPushMatrix();
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	/* do the model rotation */
	build_rotmatrix(rotmatrix, qrot);
	glMultMatrixf(&rotmatrix[0][0]);

	sscale = 1.0;

	bond_radius = BOND_RADIUS * scale;
	switch (rendermode) {
	case DM_SF:
		break;
	case DM_BS:
		sscale = bs_spheresize; /* sphere diameter */
		break;
	case DM_WF:
		sscale = 0.0;
		break;
	case DM_AT:
		bond_radius = ALPHA_BOND_RADIUS * scale;
		break;
	default:
		break;
	}
	sscale *= scale;
	dqual = DQUAL * pow(1.2, (double)quality); /* quality */

	pectr = 0;
	for (i = 0; i < atoms; i++) {
		if ((oflags & OF_IH) && (atom[i].flags & AF_H))
			continue;
		if ((oflags & OF_IW) && (atom[i].flags & AF_H2O))
			continue;

		//FIXME: this should be done using glScale really.
		patom[pectr].pa = &atom[i];
		d = atom[i].pos[V_X] - dv[V_X] * zoom;
		g = d * d;
		d = atom[i].pos[V_Y] - dv[V_Y] * zoom;
		g += d * d;
		d = atom[i].pos[V_Z] - dv[V_Z] * zoom;
		g += d * d;
		patom[pectr].r = sqrt(g);
		pectr++;
	}
//qsort(patom, pectr, 
//	      sizeof(struct adist), 
//	      (int (*)(const void *, const void *)) &cmpf);

	for (i = 0; i < pectr; i++) {
		if (rendermode == DM_AT || rendermode == DM_FAT) {
			if (NULL == patom[i].pa->achc) {
				continue;
			}
			if (patom[i].pa->achc->toward < 0) {
				if (patom[i].pa->acha->toward < 0)
                                        continue;
				pp = &acprops[2];	/* Special color for carboxyl end */
			} else if (patom[i].pa->acha->toward < 0) 
				pp = &acprops[1];	/* Special color for amino end */
			else
				pp = &acprops[0];
		} else {
			pp = &aprops[patom[i].pa->element];
		}

		glPushMatrix();
		if (rendermode == DM_FWF) {
			/* fast wireframe is special */
			glColor3f(1.0, 1.0, 1.0);
			glBegin(GL_LINES);
			for (j = 0; j < patom[i].pa->bondcount; j++) {
				glVertex3fv(patom[i].pa->pos);
				glVertex3fv(patom[patom[i].pa->bond[j].toward].pa->pos);
			}
			glEnd();
		} if (rendermode == DM_FAT) {
			/* fast alphatrace is special */
			glColor3f(1.0, 1.0, 1.0);
			glBegin(GL_LINES);
			if (patom[i].pa->achc != NULL
				&& patom[i].pa->achc->toward > 0) {
				glVertex3fv(patom[i].pa->pos);
				glVertex3fv(patom[patom[i].pa->achc->toward].pa->pos);
			}
			if (patom[i].pa->acha != NULL
				&& patom[i].pa->acha->toward > 0) {
				glVertex3fv(patom[i].pa->pos);
				glVertex3fv(patom[patom[i].pa->acha->toward].pa->pos);
			}
			glEnd();
		} else {
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, pp->color);
			glTranslatef(patom[i].pa->pos[V_X], patom[i].pa->pos[V_Y], patom[i].pa->pos[V_Z]);
			h = dqual / patom[i].r;
	
			/* render bonds */
			if (rendermode != DM_SF) {
				d = 3.0 * sqrt(h * bond_radius);
				if ((n = (int) d) < 8)
					n = 8;
				if (n > 32)
					n = 32;
				shiny = (float) n * 4.0 - 15.0;
				glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &shiny);
				if (rendermode == DM_AT) {
					if (patom[i].pa->achc != NULL && patom[i].pa->achc->toward >= 0)
						dobond(qobj, patom[i].pa->achc, n);
					if (patom[i].pa->acha != NULL && patom[i].pa->acha->toward >= 0)
						dobond(qobj, patom[i].pa->acha, n);
				} else {
					j = 0;
					for (j = 0; j < patom[i].pa->bondcount; j++) {
						dobond(qobj, &patom[i].pa->bond[j], n);
					}
				}
			}
			/* render the sphere */
			if (rendermode != DM_WF && rendermode != DM_FWF) {
				g = sscale * pp->radius;
				if ((rendermode != DM_SF) && (g < bond_radius))
					g = bond_radius;
				d = sqrt(g * h);
				if ((n = (int) d) < 10) 
					n = 10;
				if (n > 48)
					n = 48;
				shiny = (float) n * 4.0 - 15.0;
				glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &shiny);
				gluSphere(qobj, g, n, n);
			}
		}
		glPopMatrix();
	}

	glPopAttrib();
	glPopMatrix();

	free (patom);
}

#define CAM_LEFTEYE	-1
#define CAM_RIGHTEYE	1
#define CAM_CENTRE	0

int	singlecamloc = CAM_CENTRE;

float intereye = INITIAL_INTEREYE;
float depthpoint = INITIAL_DEPTHPOINT;

static void
setUpCamera(int where)
{
	float		d, e, g;

	/* This adjustment should really be in readxyz */
	e = 1.732 + LARGEST_ATOM * scale;

	d = zoom + e;
	g = sqrt(d * d + e);

	if ((d /= cos(FOV * M_PI/360.0)) < g)
		g = d;

	if ((d = zoom - e) < g * 0.01)
		d = g * 0.01;

	/* set up the camera */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(FOV / sqrt(aspect), aspect, d, g);
	if (where) {
		/* perform offset and alignment */
		glTranslatef(-intereye*where, 0, 0);
		gluLookAt(intereye*where, 0, 0,
			0, 0, -depthpoint,
			0, 1, 0);
	}

}

static void
drawScene(GLUquadricObj *qobj)
{
	/* Render the Molecule */
	glMatrixMode(GL_MODELVIEW);
	/* Clear the bitmap and render the requested image */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -zoom);

	drawMolecule(qobj);
}

/* Render an OpenGL pixmap image of the molecule */
extern void 
renderMolecule()
{
	static GLUquadricObj *	qobj = NULL;

	/* prepare the GL environment */
	initView(&qobj);

	if (oflags & OF_STEREO) {
		glDrawBuffer(GL_BACK_LEFT);
		setUpCamera(CAM_LEFTEYE);
		drawScene(qobj);

		glDrawBuffer(GL_BACK_RIGHT);
		setUpCamera(CAM_RIGHTEYE);
		drawScene(qobj);
	} else {
		glDrawBuffer(GL_BACK);
		setUpCamera(singlecamloc);
		/* Render the Molecule */
		drawScene(qobj);
	}
}
